package gassender

import (
	"context"
	"crypto/ecdsa"
	"math/big"
	"sdk_api/contract/aonUser"
	"time"

	"github.com/ethereum/go-ethereum"
	"github.com/ethereum/go-ethereum/accounts/abi/bind"
	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
	"github.com/ethereum/go-ethereum/ethclient"
	log "github.com/sirupsen/logrus"
)

type GasSender struct {
	privateKey *ecdsa.PrivateKey
	client     *ethclient.Client
	taskCh     chan *gasTask
	chainId    *big.Int
	nonce      uint64
	aonUser    *aonUser.AonUser
}

type gasTask struct {
	user    common.Address
	userId  string
	inviter string
	nonce   uint64
}

func NewGasSender(rpc, privateKey, aonUserContract string) (*GasSender, error) {
	ecdsaKey, err := crypto.HexToECDSA(common.Bytes2Hex(common.FromHex(privateKey)))
	if err != nil {
		return nil, err
	}

	client, err := ethclient.Dial(rpc)
	if err != nil {
		return nil, err
	}

	chainId, err := client.ChainID(context.Background())
	if err != nil {
		return nil, err
	}

	nonce, err := client.NonceAt(context.Background(), crypto.PubkeyToAddress(ecdsaKey.PublicKey), nil)
	if err != nil {
		return nil, err
	}

	ca, err := aonUser.NewAonUser(common.HexToAddress(aonUserContract), client)
	if err != nil {
		return nil, err
	}

	log.WithFields(log.Fields{
		"caller":       crypto.PubkeyToAddress(ecdsaKey.PublicKey),
		"userContract": common.HexToAddress(aonUserContract),
		"nonce":        nonce,
	}).Info("init gas sender")

	return &GasSender{
		privateKey: ecdsaKey,
		client:     client,
		taskCh:     make(chan *gasTask, 64),
		chainId:    chainId,
		nonce:      nonce,
		aonUser:    ca,
	}, nil

}

func (gs *GasSender) Run() {
	go func() {
		for {
			select {
			case task := <-gs.taskCh:
				task.nonce = gs.nonce
				gs.nonce++
				go gs.sendTx(task)
			}
		}
	}()
}

// func (gs *GasSender) SendAONGas(dest string, amount int) {
// 	addr := common.HexToAddress(dest)
// 	value := new(big.Int).Mul(big.NewInt(int64(amount)), big.NewInt(1000000000000000000))
// 	gs.taskCh <- &gasTask{
// 		dest:  addr,
// 		value: value,
// 	}
// }

func (gs *GasSender) AonLogin(user common.Address, userId, inviter string) {
	gs.taskCh <- &gasTask{
		user:    user,
		userId:  userId,
		inviter: inviter,
	}
}

func (gs *GasSender) sendTx(task *gasTask) {
	log.WithFields(log.Fields{
		"address":   task.user.Hex(),
		"userId":    task.userId,
		"inviterId": task.inviter,
		"nonce":     task.nonce,
	}).Info("new login task")

	opts, err := bind.NewKeyedTransactorWithChainID(gs.privateKey, gs.chainId)
	if err != nil {
		log.WithError(err).Error("create transactor failed")
		return
	}

	opts.GasPrice = big.NewInt(1000000000)
	opts.GasLimit = 5000000
	opts.Nonce = big.NewInt(int64(task.nonce))

	signedTx, err := gs.aonUser.LoginByServer(opts, task.user, task.userId, task.inviter)
	if err != nil {
		log.WithError(err).Error("send tx failed")
		return
	}
	txLog := log.WithField("txHash", signedTx.Hash().Hex())
	txLog.Info("tx broadcasted")
	for i := 0; i < 5; i++ {
		time.Sleep(time.Second * 2)
		receipt, err := gs.client.TransactionReceipt(context.Background(), signedTx.Hash())
		if err != nil && err == ethereum.NotFound {
			txLog.Info("tx receipt not found, retrying...")
			continue
		}
		if err != nil {
			txLog.WithError(err).Error("send gas tx failed")
			return
		}

		if receipt.Status != 1 {
			txLog.Error("send gas tx failed")
			return
		}
		txLog.Info("send gas tx confirmed")
		return
	}
	txLog.Error("tx receipt not found, timeout")
}
